# coding=utf-8
from collections import OrderedDict
import os
from pathlib import Path
import re
from struct import unpack
import subprocess  # nosec

from numpy import float64
from numpy import zeros

from pyaedt import generate_unique_name
from pyaedt import is_linux
from pyaedt import pyaedt_function_handler
from pyaedt import settings
from pyaedt.generic.com_parameters import COMParameters
from pyaedt.generic.general_methods import env_value
from pyaedt.misc import current_version


class SpiSim:
    """Provides support to SpiSim batch mode."""

    def __init__(self, touchstone_file=""):
        self.touchstone_file = touchstone_file
        if settings.aedt_version:
            self.desktop_install_dir = os.environ[env_value(settings.aedt_version)]
        else:
            self.desktop_install_dir = os.environ[env_value(current_version())]
        self.logger = settings.logger
        self._working_directory = ""

    @property
    def working_directory(self):
        """Working directory.

        Returns
        -------
        str
        """
        if self._working_directory != "":
            return self._working_directory
        if self.touchstone_file:
            self._working_directory = os.path.dirname(self.touchstone_file)
        return self._working_directory

    @working_directory.setter
    def working_directory(self, val):
        self._working_directory = val

    @pyaedt_function_handler()
    def _compute_spisim(self, parameter, out_file="", touchstone_file="", config_file=""):
        exec_name = "SPISimJNI_LX64.exe" if is_linux else "SPISimJNI_WIN64.exe"
        spisimExe = os.path.join(self.desktop_install_dir, "spisim", "SPISim", "modules", "ext", exec_name)

        cfgCmmd = ""
        if touchstone_file != "":
            cfgCmmd = cfgCmmd + '-i "%s"' % touchstone_file
        if config_file != "":
            cfgCmmd = '-v CFGFILE="%s"' % config_file
        if out_file:
            cfgCmmd += ' -o "%s"' % out_file
        command = [spisimExe, parameter, cfgCmmd]
        # Debug('%s %s' % (cmdList[0], ' '.join(arguments)))
        # try up to three times to be sure
        out_processing = os.path.join(out_file, generate_unique_name("spsim_out") + ".txt")
        my_env = os.environ.copy()
        my_env.update(settings.aedt_environment_variables)
        if is_linux:  # pragma: no cover
            command.append("&")
            with open(out_processing, "w") as outfile:
                subprocess.Popen(command, env=my_env, stdout=outfile, stderr=outfile).wait()  # nosec
        else:
            with open(out_processing, "w") as outfile:
                subprocess.Popen(" ".join(command), env=my_env, stdout=outfile, stderr=outfile).wait()  # nosec
        return out_processing

    @pyaedt_function_handler()
    def _get_output_parameter_from_result(self, out_file, parameter_name):
        if parameter_name == "ERL":
            try:
                with open(out_file, "r") as infile:
                    lines = infile.read()
                    parmDat = lines.split("[ParmDat]:", 1)[1]
                    for keyValu in parmDat.split(","):
                        dataAry = keyValu.split("=")
                        if dataAry[0].strip().lower() == parameter_name.lower():
                            return float(dataAry[1].strip().split()[0])
                self.logger.error(
                    "Failed to compute {}. Check input parameters and retry".format(parameter_name)
                )  # pragma: no cover
                return False  # pragma: no cover
            except IndexError:
                self.logger.error("Failed to compute {}. Check input parameters and retry".format(parameter_name))
                return False
        elif parameter_name == "COM":
            try:
                with open(out_file, "r") as infile:
                    txt = infile.read()
                com_case_0 = re.search(r"Case 0: Calculated COM = (.*?),", txt).groups()[0]
                com_case_1 = re.search(r"Case 1: Calculated COM = (.*?),", txt).groups()[0]
                return float(com_case_0), float(com_case_1)
            except IndexError:  # pragma: no cover
                self.logger.error("Failed to compute {}. Check input parameters and retry".format(parameter_name))

    @pyaedt_function_handler()
    def compute_erl(
        self,
        config_file=None,
        port_order=None,
        specify_through_ports=None,
        bandwidth=None,
        tdr_duration=None,
        z_terminations=None,
        transition_time=None,
        fixture_delay=None,
        input_amplitude=None,
        ber=None,
        pdf_bin_size=None,
        signal_loss_factor=None,
        permitted_reflection=None,
        reflections_length=None,
        modulation_type=None,
    ):
        """Compute effective return loss (ERL) using Ansys SPISIM from S-parameter file.

        Parameters
        ----------
        config_file : str, optional
            Configuration file to use as a reference. The default is ``None``, in
            which case this parameter is ignored.
        port_order : str, optional
            Whether to use "``EvenOdd``" or "``Incremental``" numbering for S4P files.
            The default is ``None``. This parameter is ignored if there are more than four ports.
        specify_through_ports : list, optional
            Input and output ports to compute the ERL on. Those are ordered like ``[inp, inneg, outp, outneg]``.
            The default is ``None``. This parameter is ignored if there are more than four ports.
        bandwidth : float, str, optional
            Application bandwidth in hertz (Hz), which is the inverse of one UI (unit interval). The value
            can be a float or a string with the unit ("m", "g"). The default is ``30e9``.
        tdr_duration : float, optional
            Time domain reflectometry (TDR) duration in seconds, meaning how long the TDR tailed data should be applied.
            The default is ``5``.
        z_terminations : float, optional
            Z-terminations (Z11 and Z22) when TDR is calculated. The default is ``50``.
        transition_time : float, str, optional
            Transition time: how fast (slew rate) input pulse transit from 0 to Vcc volt. The default is "``10p``".
        fixture_delay : float, optional
            Fixture delay: delay when input starts transition from 0 to Vcc. The default is ``500e-12``.
        input_amplitude : float, optional
            Input amplitude: Vcc volt of step input. The default is ``1.0``.
        ber : float, optional
            Specified BER: At what threshold ERL is calculated. The default is ``1e-4``.
        pdf_bin_size : float, optional
            PDF bin size: how to quantize the superimposed value. The default is ``1e-5``.
        signal_loss_factor : float, optional
            Signal loss factor (Beta). For more information, see the SPISIM Help. The default is ``1.7e9``.
        permitted_reflection : float, optional
            Permitted reflection (Rho). For more information, see the SPISIM Help. The default is ``0.18``.
        reflections_length : float, optional
            Length of the reflections: how many UI will be used to calculate ERL. The default is ``1000``.
        modulation_type : str, optional
           Modulations type: signal modulation type "``NRZ``" or "``PAM4``". The default is "``NRZ``".

        Returns
        -------
        bool or float
            Effective return loss from the spisimExe command, ``False`` when failed.
        """

        cfg_dict = {
            "INPARRY": "",
            "MIXMODE": "",
            "THRUS4P": "",
            "BANDWID": 30e9,
            "TDR_DUR": 5,
            "BINSIZE": 1e-5,
            "REFIMPD": 50,
            "SPECBER": 1e-4,
            "MODTYPE": "NRZ",
            "FIXDELY": 500e-12,
            "INPVOLT": 1.0,
            "TRSTIME": "10p",
            "SIGBETA": 1.7e9,
            "REFLRHO": 0.18,
            "NCYCLES": 1000,
        }
        if config_file:
            with open(config_file, "r") as fp:
                lines = fp.readlines()
                for line in lines:
                    if not line.startswith("#") and "=" in line:
                        split_line = [i.strip() for i in line.split("=")]
                        cfg_dict[split_line[0]] = split_line[1]

        self.touchstone_file = self.touchstone_file.replace("\\", "/")
        cfg_dict["INPARRY"] = self.touchstone_file
        cfg_dict["MIXMODE"] = "" if "MIXMODE" not in cfg_dict else cfg_dict["MIXMODE"]
        if port_order is not None and self.touchstone_file.lower().endswith(".s4p"):
            cfg_dict["MIXMODE"] = port_order
        elif not self.touchstone_file.lower().endswith(".s4p"):
            cfg_dict["MIXMODE"] = ""
        cfg_dict["THRUS4P"] = "" if "THRUS4P" not in cfg_dict else cfg_dict["THRUS4P"]

        if specify_through_ports:
            if isinstance(specify_through_ports[0], int):
                thrus4p = ",".join([str(i) for i in specify_through_ports])
            else:  # pragma: no cover
                try:
                    ports = list(self.excitations.keys())
                    thrus4p = ",".join([str(ports.index(i)) for i in specify_through_ports])
                except IndexError:
                    self.logger.error("Port not found.")
                    return False
            cfg_dict["THRUS4P"] = thrus4p

        cfg_dict["BANDWID"] = bandwidth if bandwidth is not None else cfg_dict["BANDWID"]
        cfg_dict["TDR_DUR"] = tdr_duration if tdr_duration is not None else cfg_dict["TDR_DUR"]
        cfg_dict["BINSIZE"] = pdf_bin_size if pdf_bin_size is not None else cfg_dict["BINSIZE"]
        cfg_dict["REFIMPD"] = z_terminations if z_terminations is not None else cfg_dict["REFIMPD"]
        cfg_dict["SPECBER"] = ber if ber is not None else cfg_dict["SPECBER"]
        cfg_dict["MODTYPE"] = modulation_type if modulation_type is not None else cfg_dict["MODTYPE"]
        cfg_dict["FIXDELY"] = fixture_delay if fixture_delay is not None else cfg_dict["FIXDELY"]
        cfg_dict["INPVOLT"] = input_amplitude if input_amplitude is not None else cfg_dict["INPVOLT"]
        cfg_dict["TRSTIME"] = transition_time if transition_time is not None else cfg_dict["TRSTIME"]
        cfg_dict["SIGBETA"] = signal_loss_factor if signal_loss_factor is not None else cfg_dict["SIGBETA"]
        cfg_dict["REFLRHO"] = permitted_reflection if permitted_reflection is not None else cfg_dict["REFLRHO"]
        cfg_dict["NCYCLES"] = reflections_length if reflections_length is not None else cfg_dict["NCYCLES"]

        new_cfg_file = os.path.join(self.working_directory, "spisim_erl.cfg").replace("\\", "/")
        with open(new_cfg_file, "w") as fp:
            for k, v in cfg_dict.items():
                fp.write("# {}: {}\n".format(k, k))
                fp.write("{} = {}\n".format(k, v))
        retries = 3
        trynumb = 0
        while trynumb < retries:
            out_processing = self._compute_spisim(
                "CalcERL",
                touchstone_file=self.touchstone_file,
                config_file=new_cfg_file,
                out_file=self.working_directory,
            )
            results = self._get_output_parameter_from_result(out_processing, "ERL")
            if results:
                return results
            self.logger.warning("Failing to compute ERL, retrying...")
            trynumb += 1
        self.logger.error("Failed to compute ERL.")
        return False

    @pyaedt_function_handler
    def compute_com(
        self,
        standard,
        config_file=None,
        port_order="EvenOdd",
        fext_s4p="",
        next_s4p="",
        out_folder="",
    ):
        """Compute Channel Operating Margin.

        Parameters
        ----------
        standard : str
            Name of the standard to apply.
        config_file : str, optional
            Config file to use.
        port_order : str, optional
            Whether to use "``EvenOdd``" or "``Incremental``" numbering for S4P files. The default is ``EvenOdd``.
            The default is ``None``. This parameter is ignored if there are more than four ports.
        fext_s4p : str, list, optional
            Fext touchstone file to use.
        next_s4p : str, list, optional
            Next touchstone file to use.
        out_folder : str, optional
            Output folder where to save report.

        Returns
        -------

        """
        if standard == "custom":
            com_param = COMParameters()
            com_param.load(config_file)
        else:
            com_param = COMParameters(standard)

        com_param.THRUSNP = self.touchstone_file
        com_param.FEXTARY = fext_s4p if not isinstance(fext_s4p, list) else ";".join(fext_s4p)
        com_param.NEXTARY = next_s4p if not isinstance(next_s4p, list) else ";".join(next_s4p)

        com_param.PORT_ORDER = "[1 3 2 4]" if port_order == "EvenOdd" else "[1 2 3 4]"

        com_param.RESULT_DIR = out_folder if out_folder else self.working_directory
        return self._compute_com(com_param)

    @pyaedt_function_handler
    def _compute_com(
        self,
        com_parameter,
    ):
        """Compute Channel Operating Margin.

        Parameters
        ----------
        com_parameter: :class:`COMParameters`
            COMParameters class.

        Returns
        -------

        """

        com_parameter.THRUSNP = com_parameter.THRUSNP.replace("\\", "/")
        com_parameter.FEXTARY = com_parameter.FEXTARY.replace("\\", "/")
        com_parameter.NEXTARY = com_parameter.NEXTARY.replace("\\", "/")
        com_parameter.RESULT_DIR = com_parameter.RESULT_DIR.replace("\\", "/")

        cfg_file = os.path.join(com_parameter.RESULT_DIR, "com_parameters.cfg")
        com_parameter.export(cfg_file)

        out_processing = self._compute_spisim(parameter="COM", config_file=cfg_file)
        return self._get_output_parameter_from_result(out_processing, "COM")

    @property
    def com_standards(self):
        """List of supported COM standards."""
        return list(COMParameters._STD_TABLE_MAPPING.keys())

    @staticmethod
    @pyaedt_function_handler
    def com_parameters(standard="50GAUI-1_C2C"):
        return COMParameters(standard).parameters

    @pyaedt_function_handler
    def export_com_configure_file(self, file_path, standard="50GAUI-1_C2C"):
        """Generate a configuration file for SpiSim.

        Parameters
        ----------
        file_path : str
            Full path to configuration file to create.

        Returns
        -------
        bool
        """
        return COMParameters(standard).export(file_path)


def detect_encoding(file_path, expected_pattern="", re_flags=0):
    """Check encoding of a file."""
    for encoding in ("utf-8", "utf_16_le", "cp1252", "cp1250", "shift_jis"):
        try:
            with open(file_path, "r", encoding=encoding) as f:
                lines = f.read()
                f.seek(0)
        except UnicodeDecodeError:
            # This encoding didn't work, let's try again
            continue
        else:
            if len(lines) == 0:
                # Empty file
                continue
            if expected_pattern:
                if not re.match(expected_pattern, lines, re_flags):
                    # File did not have the expected string
                    # Try again with a different encoding (This is unlikely to resolve the issue)
                    continue
            if encoding == "utf-8" and lines[1] == "\x00":
                continue
            return encoding


class DataSet(object):
    """
    This is the base class for storing all traces of a RAW file. Returned by the get_trace() or by the get_axis()
    methods.
    Normally the user doesn't have to be aware of this class. It is only used internally to encapsulate the different
    implementations of the wave population.
    Data can be retrieved directly by using the [] operator.
    If numpy is available, the numpy vector can be retrieved by using the get_wave() method.
    The parameter whattype defines what is the trace representing in the simulation, Voltage, Current a Time or
    Frequency.
    """

    def __init__(
        self,
        name,
        whattype,
        datalen,
    ):
        """Base Class for both Axis and Trace Classes.
        Defines the common operations between both."""
        self.name = name
        self.whattype = whattype
        self.data = zeros(datalen, dtype=float64)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, item):
        return self.data[item]

    @property
    def wave(self):
        """Retrieves the trace data.

        Returns
        -------
        :class:`numpy.array`
            The trace values.
        """
        return self.data


class Trace(DataSet):
    """This class is used to represent a trace.
    This class is constructed by the get_trace() command.
    If numpy is available the get_wave() method will return a numpy array.
    """

    def __init__(
        self,
        name,
        whattype,
        datalen,
        axis,
    ):
        super().__init__(name, whattype, datalen)
        self.axis = axis

    def __len__(self):
        """
        Returns the length of the axis.

        Returns
        -------
        int
            The number of data points.
        """
        return len(self.wave)


class SpiSimRawException(Exception):
    """Custom class for exception handling"""

    ...


class SpiSimRawRead(object):
    """Class for reading SPISim wave Files. It can read all types of Files."""

    @staticmethod
    def read_float64(f):
        s = f.read(8)
        return unpack("d", s)[0]

    @staticmethod
    def read_float32(f):  # pragma: no cover
        s = f.read(4)
        return unpack("f", s)[0]

    def __init__(self, raw_filename: str, **kwargs):
        raw_filename = Path(raw_filename)

        raw_file = open(raw_filename, "rb")

        ch = raw_file.read(6)
        if ch.decode(encoding="utf_8") == "Title:":
            self.encoding = "utf_8"
            sz_enc = 1
            line = "Title:"
        elif ch.decode(encoding="utf_16_le") == "Tit":  # pragma: no cover
            self.encoding = "utf_16_le"
            sz_enc = 2
            line = "Tit"
        else:  # pragma: no cover
            raise RuntimeError("Unrecognized encoding")
        settings.logger.info(f"Reading the file with encoding: '{self.encoding}' ")
        self.raw_params = OrderedDict(Filename=raw_filename)
        self.backannotations = []
        header = []
        binary_start = 6
        while True:
            ch = raw_file.read(sz_enc).decode(encoding=self.encoding, errors="replace")
            binary_start += sz_enc
            if ch == "\n":
                if self.encoding == "utf_8":
                    line = line.rstrip("\r")
                header.append(line)
                if line in ("Binary:", "Values:"):
                    self.raw_type = line
                    break
                line = ""
            else:
                line += ch

        for line in header:
            if not line.startswith("."):
                k, _, v = line.partition(":")
                if k == "Variables":
                    break
                self.raw_params[k] = v.strip()
        self.nPoints = int(self.raw_params["No. Points"], 10)
        self.nVariables = int(self.raw_params["No. Variables"], 10)
        self._traces = []

        self.axis = None
        self.flags = self.raw_params["Flags"].split()
        i = header.index("Variables:")
        ivar = 0
        for line in header[i + 1 : -1]:
            _, name, var_type = line.lstrip().split("\t")
            if ivar == 0:
                self.axis = Trace(name, var_type, self.nPoints, None)
                trace = self.axis
            else:
                trace = Trace(name, var_type, self.nPoints, self.axis)
            self._traces.append(trace)
            ivar += 1

        if len(self._traces) == 0:  # pragma: no cover
            raw_file.close()
            return

        if kwargs.get("headeronly", False):  # pragma: no cover
            raw_file.close()
            return

        if self.raw_type == "Binary:":
            for point in range(self.nPoints):
                for i, var in enumerate(self._traces):
                    value = self.read_float64(raw_file)
                    if value is not None:
                        var.data[point] = value
        else:  # pragma: no cover
            raw_file.close()
            raise SpiSimRawException("Unsupported RAW File. " "%s" "" % self.raw_type)

        raw_file.close()

        self.raw_params["No. Points"] = self.nPoints
        self.raw_params["No. Variables"] = self.nVariables
        self.raw_params["Variables"] = [var.name for var in self._traces]

    def get_raw_property(self, property_name=None):
        """
        Get a property. By default, it returns all properties defined in the RAW file.

        :param property_name: name of the property to retrieve.
        :type property_name: str
        :returns: Property object
        :rtype: str
        :raises: ValueError if the property doesn't exist
        """
        if property_name is None:
            return self.raw_params
        elif property_name in self.raw_params.keys():
            return self.raw_params[property_name]
        else:  # pragma: no cover
            raise ValueError("Invalid property. Use %s" % str(self.raw_params.keys()))

    @property
    def trace_names(self):
        """
        Returns a list of exiting trace names of the RAW file.

        Returns
        -------
        list
            Trace names.
        """
        return [trace.name for trace in self._traces]

    def get_trace(self, trace_ref):
        """Retrieves the trace with the requested name (trace_ref).

        Parameters
        ----------
        trace_ref: str, int
            Name of the trace or the index of the trace.
        """
        if isinstance(trace_ref, str):
            for trace in self._traces:
                if trace_ref.casefold() == trace.name.casefold():  # The trace names are case-insensitive
                    # assert isinstance(trace, DataSet)
                    return trace
            raise IndexError(
                f'{self} doesn\'t contain trace "{trace_ref}"\n'
                f"Valid traces are {[trc.name for trc in self._traces]}"
            )  # pragma: no cover
        else:
            return self._traces[trace_ref]

    def get_wave(self, trace_ref):
        """Retrieves the trace data with the requested name (trace_ref).

        Parameters
        ----------
        trace_ref: str, int
            Name of the trace or the index of the trace.

        Returns
        -------
        :class:`numpy.array`
            The trace values.
        """
        return self.get_trace(trace_ref).wave

    def get_axis(self):
        """This function is equivalent to get_trace(0).wave instruction.

        Returns
        -------
        :class:`numpy.array`
            Axis data.
        """
        if self.axis:
            return self.axis.wave
        else:  # pragma: no cover
            raise RuntimeError("This RAW file does not have an axis.")

    def __len__(self):
        """Compute the length of the data.

        Returns
        -------
        int
            Length of the data.
        """
        return self.axis.__len__()

    def __getitem__(self, item):
        return self.get_trace(item)



"""Miscellaneous Methods for PyAEDT."""

import os
import warnings


def list_installed_ansysem():
    """Return a list of installed AEDT versions on ``ANSYSEM_ROOT``."""
    aedt_env_var_prefix = "ANSYSEM_ROOT"
    version_list = sorted([x for x in os.environ if x.startswith(aedt_env_var_prefix)], reverse=True)
    aedt_env_var_prefix = "ANSYSEM_PY_CLIENT_ROOT"
    version_list += sorted([x for x in os.environ if x.startswith(aedt_env_var_prefix)], reverse=True)
    aedt_env_var_sv_prefix = "ANSYSEMSV_ROOT"
    version_list += sorted([x for x in os.environ if x.startswith(aedt_env_var_sv_prefix)], reverse=True)

    if not version_list:
        warnings.warn(
            "No installed versions of AEDT are found in the system environment variables ``ANSYSEM_ROOTxxx``."
        )

    return version_list


def installed_versions():
    """Get the installed AEDT versions.

    This method returns a dictionary, with version as the key and installation path
    as the value."""

    return_dict = {}
    version_list = list_installed_ansysem()
    client = False
    for version_env_var in version_list:
        if "ANSYSEMSV_ROOT" in version_env_var:
            current_version_id = version_env_var.replace("ANSYSEMSV_ROOT", "")
            student = True
        elif "ANSYSEM_ROOT" in version_env_var:
            current_version_id = version_env_var.replace("ANSYSEM_ROOT", "")
            student = False
        else:
            current_version_id = version_env_var.replace("ANSYSEM_PY_CLIENT_ROOT", "")
            student = False
            client = True
        try:
            version = int(current_version_id[0:2])
            release = int(current_version_id[2])
            if version < 20:
                if release < 3:
                    version -= 1
                else:
                    release -= 2
            if student:
                v_key = "20{0}.{1}SV".format(version, release)
            elif client:
                v_key = "20{0}.{1}CL".format(version, release)
            else:
                v_key = "20{0}.{1}".format(version, release)
            return_dict[v_key] = os.environ[version_env_var]
        except:  # pragma: no cover
            pass
    return return_dict


def current_version():
    """Get the current AEDT version."""
    try:
        return list(installed_versions().keys())[0]
    except (NameError, IndexError):
        return ""


def current_student_version():
    """Get the current AEDT student version."""
    for version_key in installed_versions():
        if "SV" in version_key:
            return version_key
    return ""



from pyedb import Edb as App

from pyaedt.generic.settings import settings

log = settings.logger

log.warning("pyaedt.edb.Edb is Deprecated. Please use pyedb.Edb or pyaedt.Edb call instead")
Edb = App



import json
import os
import re
import traceback

import pyaedt
from pyaedt.aedt_logger import pyaedt_logger as logger
from pyaedt.generic.general_methods import check_and_download_file
from pyaedt.generic.general_methods import check_if_path_exists
from pyaedt.generic.settings import settings


class Component:
    """Component extracted from ibis model."""

    def __init__(self):
        self._name = None
        self._manufacturer = None
        self._pins = {}

    @property
    def name(self):
        """Name of the component.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].name
        'MT47H64M4BP-3_25'

        """

        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def manufacturer(self):
        """Manufacturer of the component.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, 'u26a_800_modified.ibs'), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].manufacturer
        'Micron Technology, Inc.'

        """

        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value):
        self._manufacturer = value

    @property
    def pins(self):
        """Pins of the component.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> pins = ibis.components["MT47H64M4BP-3_25"].pins

        """

        return self._pins

    @pins.setter
    def pins(self, value):
        self._pins = value


class Pin:
    """Pin from a component with all its data feature.

    Parameters
    ----------
    name : str
        Name of the pin.
    circuit : class:`pyaedt.circuit.Circuit`
        Circuit in which the pin will be added to.
    """

    def __init__(self, name, buffername, circuit):
        self._name = name
        self._buffer_name = buffername
        self._circuit = circuit
        self._short_name = None
        self._signal = None
        self._model = None
        self._r_value = None
        self._l_value = None
        self._c_value = None

    @property
    def name(self):
        """Full name of the pin including the component name and the ibis filename.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].name
        'A1_MT47H64M4BP-3_25_u26a_800'

        """
        return self._name

    @property
    def buffer_name(self):
        """Full name of the buffer including the component name and the ibis filename."""
        return self._buffer_name

    @property
    def short_name(self):
        """Name of the pin without the name of the component.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].short_name
        'A1'

        """
        return self._short_name

    @short_name.setter
    def short_name(self, value):
        self._short_name = value

    @property
    def signal(self):
        """Signal of the pin.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].signal
        'VDD'

        """
        return self._signal

    @signal.setter
    def signal(self, value):
        self._signal = value

    @property
    def model(self):
        """Model of the pin.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].signal
        'POWER'

        """
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def r_value(self):
        """Resistance value in ohms.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].r_value
        '44.3m'

        """

        return self._r_value

    @r_value.setter
    def r_value(self, value):
        self._r_value = value

    @property
    def l_value(self):
        """Inductance value in H.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].l_value
        '1.99nH'

        """
        return self._l_value

    @l_value.setter
    def l_value(self, value):
        self._l_value = value

    @property
    def c_value(self):
        """Capacitance value in F.

        Examples
        --------
        >>> ibis = ibis_reader.IbisReader(os.path.join(path_to_ibis_files, "u26a_800_modified.ibs"), circuit)
        >>> ibis.components["MT47H64M4BP-3_25"].pins["A1_MT47H64M4BP-3_25_u26a_800"].c_value
        '0.59pF'

        """
        return self._c_value

    @c_value.setter
    def c_value(self, value):
        self._c_value = value

    def add(self):
        """Add a pin to the list of components in the Project Manager."""
        try:
            return self._circuit.modeler.schematic.o_component_manager.AddSolverOnDemandModel(
                self.buffer_name,
                [
                    "NAME:CosimDefinition",
                    "CosimulatorType:=",
                    7,
                    "CosimDefName:=",
                    "DefaultIBISNetlist",
                    "IsDefinition:=",
                    True,
                    "Connect:=",
                    True,
                    "Data:=",
                    [],
                    "GRef:=",
                    [],
                ],
            )
        except:
            logger.error("Error adding {} pin component.".format(self.short_name))
            return False

    def insert(self, x, y, angle=0.0):
        """Insert a pin at a defined location inside the graphical window.

        Parameters
        ----------
        x : float
            X position of the pin.
        y : float
            Y position of the pin.
        angle : float, optional
            Angle of the pin. The default value is ``"0.0"``.

        Returns
        -------
        :class:`pyaedt.modeler.Object3d.CircuitComponent`
            Circuit Component Object.

        """

        return self._circuit.modeler.schematic.create_component(
            component_library=None,
            component_name=self.buffer_name,
            location=[x, y],
            angle=angle,
        )


class Buffer:
    def __init__(self, ibis_name, short_name, circuit):
        self._ibis_name = ibis_name
        self._short_name = short_name
        self._circuit = circuit

    @property
    def name(self):
        """Full name of the buffer including the ibis filename."""
        return "{}_{}".format(self.short_name, self._ibis_name)

    @property
    def short_name(self):
        """Short name of the buffer without the ibis filename included."""
        return self._short_name

    def add(self):
        """Add a buffer to the list of components in the Project Manager."""
        self._circuit.modeler.schematic.o_component_manager.AddSolverOnDemandModel(
            self.name,
            [
                "NAME:CosimDefinition",
                "CosimulatorType:=",
                7,
                "CosimDefName:=",
                "DefaultIBISNetlist",
                "IsDefinition:=",
                True,
                "Connect:=",
                True,
                "Data:=",
                [],
                "GRef:=",
                [],
            ],
        )

    def insert(self, x, y, angle=0.0):
        """Insert a buffer at a defined location inside the graphical window.

        Parameters
        ----------
        x : float
            X position of the buffer.
        y : float
            Y position of the buffer.
        angle : float, optional
            Angle of the buffer. The default value is ``"0.0"``.

        Returns
        -------
        :class:`pyaedt.modeler.Object3d.CircuitComponent`
            Circuit Component Object.

        """

        return self._circuit.modeler.schematic.create_component(
            component_library=None,
            component_name=self.name,
            location=[x, y],
            angle=angle,
        )


class ModelSelector:
    def __init__(self):
        self._model_selector_items = []
        self._name = None

    @property
    def model_selector_items(self):
        """Model selector items."""
        return self._model_selector_items

    @model_selector_items.setter
    def model_selector_items(self, value):
        self._model_selector_items = value

    @property
    def name(self):
        """Name of the model selector."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


class ModelSelectorItem:
    def __init__(self):
        self._description = []
        self._name = None

    @property
    def description(self):
        """Description of the item."""
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def name(self):
        """Name of the item."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


class Model:
    def __init__(self):
        self._description = []
        self._name = None
        self._clamp = None
        self._enable = None
        self._ami = None
        self._c_comp = None

    @property
    def name(self):
        """Name of the item."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def model_type(self):
        """Type of the model."""
        return self._model_type

    @model_type.setter
    def model_type(self, value):
        self._model_type = value

    @property
    def clamp(self):
        """Clamp."""
        return self._clamp

    @clamp.setter
    def clamp(self, value):
        self._clamp = value

    @property
    def enable(self):
        """Is model enabled or not."""
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = value

    @property
    def ami(self):
        """Is model enabled or not."""
        return self._ami

    @ami.setter
    def ami(self, value):
        self._ami = value

    @property
    def c_comp(self):
        """Is model enabled or not."""
        return self._c_comp

    @c_comp.setter
    def c_comp(self, value):
        self._c_comp = value


class Ibis:
    """Ibis model with all data extracted: name, components, models.

    Parameters
    ----------
    name : str
        Name of ibis model.
    circuit : class:`pyaedt.circuit.Circuit`
        Circuit in which the ibis components will be used.
    """

    # Ibis reader must work independently or in Circuit.
    def __init__(self, name, circuit):
        self.circuit = circuit
        self._name = name
        self._components = {}
        self._model_selectors = []
        self._models = []

    @property
    def name(self):
        """Name of the ibis model."""
        return self._name

    @property
    def components(self):
        """List of all components included in the ibis file."""
        return self._components

    @components.setter
    def components(self, value):
        self._components = value

    @property
    def model_selectors(self):
        """List of all model selectors included in the ibis file."""
        return self._model_selectors

    @model_selectors.setter
    def model_selectors(self, value):
        self._model_selectors = value

    @property
    def models(self):
        """List of all models included in the ibis file."""
        return self._models

    @models.setter
    def models(self, value):
        self._models = value

    @property
    def buffers(self):
        """Buffers included into the ibis model."""
        return self._buffers

    @buffers.setter
    def buffers(self, value):
        self._buffers = value


class AMI:
    """Ibis-AMI model with all data extracted: name, components, models.

    Parameters
    ----------
    name : str
        Name of ibis model.
    circuit : class:`pyaedt.circuit.Circuit`
        Circuit in which the ibis components will be used.
    """

    # Ibis reader must work independently or in Circuit.
    def __init__(self, name, circuit):
        self.circuit = circuit
        self._name = name
        self._components = {}
        self._model_selectors = []
        self._models = []

    @property
    def name(self):
        """Name of the ibis model."""
        return self._name

    @property
    def components(self):
        """List of all components included in the ibis file."""
        return self._components

    @components.setter
    def components(self, value):
        self._components = value

    @property
    def model_selectors(self):
        """List of all model selectors included in the ibis file."""
        return self._model_selectors

    @model_selectors.setter
    def model_selectors(self, value):
        self._model_selectors = value

    @property
    def models(self):
        """List of all models included in the ibis file."""
        return self._models

    @models.setter
    def models(self, value):
        self._models = value

    @property
    def buffers(self):
        """Buffers included into the ibis model."""
        return self._buffers

    @buffers.setter
    def buffers(self, value):
        self._buffers = value


class IbisReader(object):
    """Reads *.ibis file content.
    Setup an Ibis object exposing all the extracted data.

    Parameters
    ----------
    filename : str
        Name of ibis model.
    circuit : class:`pyaedt.circuit.Circuit`
        Circuit in which the ibis components will be used.
    """

    def __init__(self, filename, circuit):
        self._filename = filename
        self._circuit = circuit
        self._ibis_model = None

    @property
    def ibis_model(self):
        "Ibis model gathering the entire set of data extracted from the \\*.ibis file."
        return self._ibis_model

    def parse_ibis_file(self):
        """Reads \\*.ibis file content.

        Parameters
        ----------
        filename : str
            Name of ibis model.
        circuit : class:`pyaedt.circuit.Circuit`
            Circuit in which the ibis components will be used.

        Returns
        ----------
        :class:`pyaedt.generic.ibis_reader.Ibis`
            Ibis object exposing all data from the ibis file.

        Examples
        --------
        Read u26a_800.ibs file provided in the AEDT suit installation.
        >>> import os
        >>> from pyaedt import Desktop
        >>> from pyaedt.circuit import Circuit
        >>> from pyaedt.generic import ibis_reader
        >>> desktop = Desktop()
        >>> circuit = Circuit()
        >>> ibis = ibis_reader.IbisReader(os.path.join(desktop.install_path, "buflib", "IBIS", "u26a_800.ibs"), circuit)

        """

        if not check_if_path_exists(self._filename):
            raise Exception("{} does not exist.".format(self._filename))

        ibis_name = pyaedt.generic.general_methods.get_filename_without_extension(self._filename)
        ibis = Ibis(ibis_name, self._circuit)
        if settings.remote_rpc_session_temp_folder:
            local_path = os.path.join(settings.remote_rpc_session_temp_folder, os.path.split(self._filename)[-1])
            file_to_open = check_and_download_file(local_path, self._filename)
        else:
            file_to_open = self._filename

        # Read *.ibis file.
        ibis_info = ibis_parsing(self._filename)
        component_selector = [ibis_info[item] for item in ibis_info if "component" in item]

        self.read_component(ibis, component_selector)

        model_selector = [ibis_info[item] for item in ibis_info if "model selector" in item]
        self.read_model_selector(ibis, model_selector)

        # model = [ibis_info[item] for item in ibis_info if 'selector' not in item and 'model' in item]
        model = [ibis_info[item] for item in ibis_info if re.match(r"^model\d*$", item) is not None]

        self.read_model(ibis, model)

        buffers = {}
        for model_selector in ibis.model_selectors:
            buffer = Buffer(ibis_name, model_selector.name, self._circuit)
            buffers[buffer.name] = buffer

        for model in ibis.models:
            buffer = Buffer(ibis_name, model.name, self._circuit)
            buffers[buffer.name] = buffer

        ibis.buffers = buffers

        if self._circuit:
            args = [
                "NAME:Options",
                "Mode:=",
                4,
                "Overwrite:=",
                False,
                "SupportsSimModels:=",
                False,
                "LoadOnly:=",
                False,
            ]
            arg_buffers = ["NAME:Buffers"]
            for buffer_item in buffers.values():
                arg_buffers.append("{}:=".format(buffer_item.short_name))
                arg_buffers.append([True, "IbisSingleEnded"])
            model_selector_names = [i.name for i in ibis.model_selectors]
            arg_components = ["NAME:Components"]
            for comp_value in ibis.components.values():
                arg_component = ["NAME:{}".format(comp_value.name)]
                for pin in comp_value.pins.values():
                    arg_component.append("{}:=".format(pin.short_name))
                    if pin.model not in model_selector_names:
                        arg_component.append([False, False])
                    else:
                        arg_component.append([True, False])
                arg_components.append(arg_component)

            args.append(arg_buffers)
            args.append(arg_components)

            self._circuit.modeler.schematic.o_component_manager.ImportModelsFromFile(self._filename, args)

        self._ibis_model = ibis
        return ibis_info

    # Model
    def read_model(self, ibis, model_list):
        """Extracts model's info.

        Parameters
        ----------
        ibis : :class:`pyaedt.generic.ibis_reader.Ibis`
            ibis object containing all info.
        current_line : str
            Current line content.
        ibis_file : TextIO
            File's stream.

        """
        for model_info in model_list:
            model_spec_info = model_info["model"].strip().split("\n")
            for idx, model_spec in enumerate(model_spec_info):
                if not idx:
                    model = Model()
                    model.name = model_spec
                else:
                    if is_started_with(model_spec.lower(), "model_type"):
                        # model.model_type = model_spec.split()[-1].strip()
                        model.model_type = model_spec.split()[-1].strip()
                        # iStart = model_spec.index(" ", 1)
                        # if iStart > 0:
                        #     model.ModelType = model_spec[iStart:].strip()
                    elif is_started_with(model_spec.lower(), "c_comp"):
                        model.c_comp = model_spec.split()[1:]

                    elif is_started_with(model_spec.lower(), "enable ", True):
                        model.enable = model_spec.split()[-1].strip()

            model_info_lower = {key.lower(): value for key, value in model_info.items()}

            if "gnd clamp" in [key.lower() for key in model_info.keys()]:
                model.clamp = True
            if "algorithmic model" in [key.lower() for key in model_info.keys()]:
                matching_key = next((key for key in model_info.keys() if "algorithmic model" in key.lower()), None)
                ami_info = model_info[matching_key][matching_key].split()
                model.ami = model_info[matching_key][matching_key].split()
                ibis.AMI = True
            else:
                ibis.AMI = False

            ibis.models.append(model)

    # Model Selector
    def read_model_selector(self, ibis, model_selector_list):
        """Extracts model selector's info.

        Parameters
        ----------
        ibis : :class:`pyaedt.generic.ibis_reader.Ibis`
            ibis object containing all info.
        current_line : str
            Current line content.
        ibis_file : TextIO
            File's stream.

        """

        for model_selector_info in model_selector_list:
            model_selector_info = model_selector_info["model selector"].strip().split("\n")
            for idx, model in enumerate(model_selector_info):
                if not idx:
                    model_selector = ModelSelector()
                    model_selector.model_selector_items = []
                    model_selector.name = model
                else:
                    model_selector.model_selector_items.append(self.make_model(model.strip()))

            # ModelSelectorItem
            ibis.model_selectors.append(model_selector)

    @classmethod
    def make_model(cls, current_line):
        """Creates model object.

        Parameters
        ----------
        current_line : str
            Current line content.

        Returns
        -------
        :class:`pyaedt.generic.ibis_reader.ModelSelectorItem`
            Model selector item.

        """

        item = ModelSelectorItem()
        i_start = current_line.index(" ", 1)

        if i_start > 0:
            item.name = current_line[:i_start].strip()
            item.description = current_line[i_start:].strip()

        return item

    # Component
    def read_component(self, ibis, comp_infos):
        """Extracts component's info.

        Parameters
        ----------
        ibis : :class:`pyaedt.generic.ibis_reader.Ibis`
            ibis object containing all info.
        comp_infos : list

        """
        if not isinstance(comp_infos, list):
            comp_infos = [comp_infos]
        for comp_info in comp_infos:
            component = Component()
            component.name = comp_info["component"]
            component.manufacturer = comp_info["manufacturer"]["manufacturer"]
            self.fill_package_info(component, comp_info["package"]["package"])
            pin_list = comp_info["pin"]["pin"].strip().split("\n")[1:]
            for pin_info in pin_list:
                pin = self.make_pin_object(pin_info, component.name, ibis)
                component.pins[pin.name] = pin

            ibis.components[component.name] = component

    @classmethod
    def fill_package_info(cls, component, pkg_info):
        """Extracts model's info.

        Parameters
        ----------
        component : :class:`pyaedt.generic.ibis_reader.Component`
            Current line content.
        current_line : str
            Current line content.
        ibis_file : TextIO
            File's stream.

        """
        pkg_info = pkg_info.strip().split("\n")
        for rlc in pkg_info:
            if is_started_with(rlc, "R_pkg"):
                component.R_pkg = rlc.strip()
            elif is_started_with(rlc, "L_pkg"):
                component.L_pkg = rlc.strip()
            elif is_started_with(rlc, "C_pkg"):
                component.C_pkg = rlc.strip()

    @classmethod
    def get_component_name(cls, line):
        """Gets the name of the component.

        Parameters
        ----------
        line : str
            Current line content.

        Returns
        -------
        str
            Name of the component.

        """
        return line.replace("[Component]", "").strip()

    def make_pin_object(self, line, component_name, ibis):
        """Extracts model's info.

        Parameters
        ----------
        line : str
            Current line content.
        component_name : str
            Name of the component.
        ibis : :class:`pyaedt.generic.ibis_reader.Ibis`
            ibis object containing all info.

        Returns
        -------
        :class:`pyaedt.generic.ibis_reader.Pin`
            Pin object.

        """

        current_string = ""

        current_string = line.strip().replace("\t", " ")

        pin_name = self.get_first_parameter(current_string)
        current_string = current_string[len(pin_name) + 1 :].strip()

        signal = self.get_first_parameter(current_string)
        current_string = current_string[len(signal) + 1 :].strip()

        model = self.get_first_parameter(current_string)
        current_string = current_string[len(model) + 1 :].strip()

        r_value = self.get_first_parameter(current_string)
        current_string = current_string[len(r_value) + 1 :].strip()

        l_value = self.get_first_parameter(current_string)
        current_string = current_string[len(l_value) + 1 :].strip()

        c_value = self.get_first_parameter(current_string)

        pin = Pin(
            pin_name + "_" + component_name + "_" + ibis.name,
            signal + "_" + component_name + "_" + ibis.name,
            ibis.circuit,
        )
        pin.short_name = pin_name
        pin.signal = signal
        pin.model = model
        pin.r_value = r_value
        pin.l_value = l_value
        pin.c_value = c_value

        return pin

    @classmethod
    def get_first_parameter(cls, line):
        """Gets first parameter string value.

        Parameters
        ----------
        line : str
            Current line content.

        Returns
        -------
        str
            First info extracted from the current line.

        """

        if line == "":
            return ""

        data = line.split(" ")
        return data[0].strip()


class AMIReader(IbisReader):
    """Reads *.ibis file content.
    Setup an Ibis object exposing all the extracted data.

    Parameters
    ----------
    filename : str
        Name of ibis model.
    circuit : class:`pyaedt.circuit.Circuit`
        Circuit in which the ibis components will be used.
    """

    def __init__(self, filename, circuit):
        self._filename = filename
        self._circuit = circuit
        self._ami_model = None

    @property
    def ami_model(self):
        "Ibis-AMI model gathering the entire set of data extracted from the \\*.ami file."
        return self._ami_model

    def parse_ibis_file(self):
        """Reads \\*.ami file content.

        Parameters
        ----------
        filename : str
            Name of ibis model.
        circuit : class:`pyaedt.circuit.Circuit`
            Circuit in which the ibis components will be used.

        Returns
        ----------
        :class:`pyaedt.generic.ibis_reader.Ibis`
            Ibis object exposing all data from the ibis file.

        Examples
        --------
        Read u26a_800.ibs file provided in the AEDT suit installation.
        >>> import os
        >>> from pyaedt import Desktop
        >>> from pyaedt.circuit import Circuit
        >>> from pyaedt.generic import ibis_reader
        >>> desktop = Desktop()
        >>> circuit = Circuit()
        >>> ibis = ibis_reader.IbisReader(os.path.join(desktop.install_path, "buflib", "IBIS", "u26a_800.ibs"), circuit)
        """

        if not check_if_path_exists(self._filename):
            raise Exception("{} does not exist.".format(self._filename))

        ami_name = pyaedt.generic.general_methods.get_filename_without_extension(self._filename)
        ibis = AMI(ami_name, self._circuit)
        if settings.remote_rpc_session_temp_folder:
            local_path = os.path.join(settings.remote_rpc_session_temp_folder, os.path.split(self._filename)[-1])
            file_to_open = check_and_download_file(local_path, self._filename)
        else:
            file_to_open = self._filename

        # Read *.ibis file.
        ibis_info = ibis_parsing(self._filename)
        component_selector = [ibis_info[item] for item in ibis_info if "component" in item]

        self.read_component(ibis, component_selector)

        model_selector = [ibis_info[item] for item in ibis_info if "model selector" in item]
        self.read_model_selector(ibis, model_selector)

        # model = [ibis_info[item] for item in ibis_info if 'selector' not in item and 'model' in item]
        model = [ibis_info[item] for item in ibis_info if item.startswith("model")]

        self.read_model(ibis, model)

        buffers = {}
        for model_selector in ibis.model_selectors:
            buffer = Buffer(ami_name, model_selector.name, self._circuit)
            buffers[buffer.name] = buffer

        for model in ibis.models:
            buffer = Buffer(ami_name, model.name, self._circuit)
            buffers[buffer.name] = buffer

        ibis.buffers = buffers

        if self._circuit:
            args = [
                "NAME:Options",
                "Mode:=",
                4,
                "Overwrite:=",
                False,
                "SupportsSimModels:=",
                False,
                "LoadOnly:=",
                False,
            ]
            arg_buffers = ["NAME:Buffers"]
            for buffer in buffers:
                arg_buffers.append("{}:=".format(buffers[buffer].short_name))
                arg_buffers.append([True, "IbisSingleEnded"])
            model_selector_names = [i.name for i in ibis.model_selectors]
            arg_components = ["NAME:Components"]
            for component in ibis.components:
                arg_component = ["NAME:{}".format(ibis.components[component].name)]
                for pin in ibis.components[component].pins:
                    arg_component.append("{}:=".format(ibis.components[component].pins[pin].short_name))
                    if ibis.components[component].pins[pin].model not in model_selector_names:
                        arg_component.append([False, False])
                    else:
                        arg_component.append([True, False])
                arg_components.append(arg_component)

            args.append(arg_buffers)
            args.append(arg_components)

            self._circuit.modeler.schematic.o_component_manager.ImportModelsFromFile(self._filename, args)

        self._ibis_model = ibis
        return ibis_info


def is_started_with(src, find, ignore_case=True):
    """Verifies if a string content starts with a specific string or not.

    This is identical to ``str.startswith``, except that it includes
    the ``ignore_case`` parameter.

    Parameters
    ----------
    src : str
        Current line content.
    find : str
        Current line content.
    ignore_case : bool, optional
        Case sensitive or not. The default value is ``True``.

    Returns
    -------
    bool
        ``True`` if the src string starts with the pattern.

    """

    if ignore_case:
        return src.lower().startswith(find.lower())
    return src.startswith(find)


def lowercase_json(json_data):
    """Convert a json structure to lower case."""
    if isinstance(json_data, str):
        return json_data.lower()
    elif isinstance(json_data, dict):
        return {lowercase_json(k): lowercase_json(v) for k, v in json_data.items()}
    elif isinstance(json_data, list):
        return [lowercase_json(item) for item in json_data]
    else:
        return json_data


def ibis_parsing(file):
    """Open and parse ibis file using json Ibis template.

    Parameters
    ----------
    file : str
        File name to parse.
    """
    ibis = {}
    # OPEN AND READ IBIS FILE
    with open(file, "r") as fp:
        ibis_data = list(enumerate(fp))

    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "ibis_v7.json"), "r") as f:
        ibis_ref = json.load(f)
    ibis_ref = lowercase_json(ibis_ref)

    # FOR EACH LINE
    try:
        level = -1
        key_iter = [0, 0, 0, 0]
        pre_key_ref = ["", "", "", ""]
        pre_key_save = ["", "", "", ""]
        for idx, line in ibis_data:
            # COMMENT
            if line[0] == "|":
                pass

            # KEYWORD START
            elif line[0] == "[":
                # FIND IBIS KEYWORD : [keyword]
                key = line.split("[")[-1].split("]")[0].replace("_", " ")
                key_ref = key.lower()
                # if key == 'End':
                #     print("JH")
                val = line.split("]")[-1].strip()
                if key_ref == "model":
                    pass

                if "end" in key_ref:
                    pass

                # FOR TOP LEVEL KEYWORD
                elif key_ref in ibis_ref.keys():
                    level = 0
                    if key_ref in ibis.keys():
                        key_iter[0] += 1
                        key_save = key_ref + str(key_iter[0])
                    else:
                        key_save = key_ref

                    ibis[key_save] = {}
                    ibis[key_save][key_ref] = val
                    pre_key_ref[0] = key_ref
                    pre_key_save[0] = key_save

                # FOR 2ND LEVEL KEYWORD
                elif key_ref in ibis_ref[pre_key_ref[0]].keys():
                    level = 1
                    if key_ref in ibis[pre_key_save[0]].keys():
                        key_iter[1] += 1
                        key_save = key_ref + str(key_iter[1])
                    else:
                        key_save = key_ref
                    ibis[pre_key_save[0]][key_save] = {}
                    ibis[pre_key_save[0]][key_save][key_ref] = val
                    pre_key_ref[1] = key_ref
                    pre_key_save[1] = key_save

                # FOR third LEVEL KEYWORD
                elif key_ref in ibis_ref[pre_key_ref[0]][pre_key_ref[1]].keys():
                    level = 2
                    if key_ref in ibis[pre_key_save[0]][pre_key_save[1]].keys():
                        key_iter[2] += 1
                        key_save = key_ref + str(key_iter[2])
                    else:
                        key_save = key_ref
                    ibis[pre_key_save[0]][pre_key_save[1]][key_save] = {}
                    ibis[pre_key_save[0]][pre_key_save[1]][key_save][key_ref] = val
                    pre_key_ref[2] = key_ref
                    pre_key_save[2] = key_save

                # FOR 4TH LEVEL KEYWORD
                elif key_ref in ibis_ref[pre_key_ref[0]][pre_key_ref[1]][pre_key_ref[2]].keys():
                    level = 3
                    if key_ref in ibis[pre_key_save[0]][pre_key_save[1]][pre_key_save[2]].keys():
                        key_iter[3] += 1
                        key_save = key_ref + str(key_iter[3])
                    else:
                        key_save = key_ref
                    ibis[pre_key_save[0]][pre_key_save[1]][pre_key_save[2]][key_save] = {}
                    ibis[pre_key_save[0]][pre_key_save[1]][pre_key_save[2]][key_save][key_ref] = val
                    pre_key_ref[3] = key_ref
                    pre_key_save[3] = key_save

                else:
                    logger.error("Invalid IBIS Keyword : {}".format(key))
                    return False

            # ALREADY FIND OUT KEYWORD
            else:
                # IF NOT BLANK LINE
                if not line.strip() == "":
                    # FOR TOP LEVEL KEYWORD
                    if level == 0:
                        ibis[pre_key_save[0]][key_ref] += "\n" + line.strip()

                    elif level == 1:
                        ibis[pre_key_save[0]][pre_key_save[1]][key_ref] += "\n" + line.strip()

                    elif level == 2:
                        ibis[pre_key_save[0]][pre_key_save[1]][pre_key_save[2]][key_ref] += "\n" + line.strip()

                    elif level == 3:
                        ibis[pre_key_save[0]][pre_key_save[1]][pre_key_save[2]][pre_key_save[3]][key_ref] += (
                            "\n" + line.strip()
                        )
    except Exception:
        logger.error(traceback.format_exc())
        return False

    # RETURN IBIS PARSING RESULT
    return ibis



def get_hpc_info(filename):
    """Retrieve HPC information.

    Parameters
    ----------
    filename : str
        Name of the file.

    Returns
    -------
    config_name : str
        Config name.
    design_type : str
        Design name.

    """
    config_name = ""
    design_type = ""
    with open(filename, "r") as fid:
        for line in fid:
            if "ConfigName=" in line:
                config_name = line.strip().replace("ConfigName=", "").replace("'", "")
            elif "DesignType=" in line:
                design_type = line.strip().replace("DesignType=", "").replace("'", "")
    return config_name, design_type


def update_hpc_option(filnename, propertyname, propertyvalue, isvaluestring=True, separator="="):
    """Update an HPC option in the configuration file.

    Parameters
    ----------
    filnename : str
        Full path and name of the configuration file. The file type can be ACF or TXT.
    propertyname : str
        Name of the property to update.
    propertyvalue :
        Value for the property.
    isvaluestring : bool, optional
        Whether the value  is a string. The default is ``True``.
    separator : str, optional
        Separates the property name from its value. The default is ``=``.

    Returns
    -------
    type

    """
    line_number = None
    new_line = ""
    with open(filnename, "r") as file:
        data = file.readlines()
        for i, line in enumerate(data):
            if propertyname + separator in line:
                old_prop = line.strip()
                old_line = line
                line_number = i
                if isvaluestring:
                    new_line = old_line.replace(old_prop, propertyname + separator + "'" + str(propertyvalue) + "'")
                else:
                    new_line = old_line.replace(old_prop, propertyname + separator + str(propertyvalue))
                break

    data[line_number] = new_line
    try:
        with open(filnename, "w") as file:
            file.writelines(data)
        return True
    except IOError:
        return False


def update_simulation_cores(name, nc):
    """Update the HPC number of cores in the configuration file.

    Parameters
    ----------
    name : str
        Name of the configuration file.
    nc : int or string
        Number of simulation cores.

    """
    with open(name) as fid:
        for line in fid:
            if "NumCores=" in line:
                old_cores = line.strip()
                new_line = line
    with open(name) as fid:
        new_line = fid.read().replace(old_cores, "NumCores=" + str(nc))
    with open(name, "w") as f:
        f.write(new_line)


def update_simulation_engines(name, nc):
    """Update the HPC number of simulaton engines in the configuration file.

    Parameters
    ----------
    name : str
        Name of the configuration file.
    nc : int or str
        Number of simulaton engines.

    """
    with open(name) as fid:
        for line in fid:
            if "NumEngines=" in line:
                old_cores = line.strip()
                new_line = line
    with open(name) as fid:
        new_line = fid.read().replace(old_cores, "NumEngines=" + str(nc))
    with open(name, "w") as f:
        f.write(new_line)


def update_machine_name(name, machinename):
    """Update the machine name.

    Parameters
    ----------
    name : str
        Path of the configuration file.
    machinename : str
        New name of the machine.

    """
    with open(name) as fid:
        for line in fid:
            if "MachineName=" in line:
                old_machine = line.strip()
                new_line = line
    with open(name) as fid:
        new_line = fid.read().replace(old_machine, "MachineName='" + str(machinename) + "'")
    with open(name, "w") as f:
        f.write(new_line)


def update_config_name(name, machinename):
    """Update the name of the configuration.

    Parameters
    ----------
    name : str
        Path of the configuration file.
    machinename : str
        New name of the machine.

    """
    with open(name) as fid:
        for line in fid:
            if "ConfigName=" in line:
                old_config = line.strip()
                new_line = line
    with open(name) as fid:
        if machinename == "localhost":
            machinename = "Local"
        new_line = fid.read().replace(old_config, "ConfigName='" + str(machinename) + "'")
    with open(name, "w") as f:
        f.write(new_line)


def update_cluster_cores(file_name, param_name, param_val):
    """Update the number of cluster cores in the configuration file.

    Parameters
    ----------
    file_name : str
        Full path and name of the configuration file. The file type can be ACF or TXT.
    param_name : str
        Name of the parameter.
    param_val : int
         New number of cluster cores.

    """
    with open(file_name) as f:
        for line in f:
            if param_name in line:
                old_line = line
    with open(file_name) as f:
        replacement_line = "\\	\\	\\	NumCores=" + str(param_val) + "\\" + chr(10)
        new_line = f.read().replace(old_line, replacement_line)
    with open(file_name, "w") as f:
        f.write(new_line)


def update_hpc_template(file_name, param_name, param_val):
    """Update a parameter in the HPC template file.

    Parameters
    ----------
    file_name : str
        Full path and name of the HPC template file.
    param_name : str
        Name of the parameter to update.
    param_val : int
        Value of the parameter.

    """
    with open(file_name) as f:
        for line in f:
            if line.find(param_name) > 0:
                line = f.readline()
                old_line = line
    with open(file_name) as f:
        replacement_line = "\\	\\	\\	Value=\\'" + str(param_val) + "\\'\\" + chr(10)
        new_line = f.read().replace(old_line, replacement_line)
    with open(file_name, "w") as f:
        f.write(new_line)



"""
Enums from the ``EmitApiPython`` module are defined as ``None`` until this module initializes.
This allows IDE auto-complete to find them and ``emit_constants`` to import before the
``EmitApiPython`` module has loaded (typically when a ``pyaedt.Emit()`` object is created).

Because the members must be reassigned at runtime, the Enum class cannot be used.
"""


class MutableEnum:
    @classmethod
    def members(cls):
        members = [
            getattr(cls, attr) for attr in dir(cls) if not callable(getattr(cls, attr)) and not attr.startswith("__")
        ]
        if members[0] is None:
            raise Exception(
                "Enum is uninitialized. Create a pyaedt.Emit() object before using {}.".format(cls.__name__)
            )
        return members


class ResultType(MutableEnum):
    EMI = None
    DESENSE = None
    SENSITIVITY = None
    POWER_AT_RX = None


class TxRxMode(MutableEnum):
    TX = None
    RX = None
    BOTH = None


class InterfererType(MutableEnum):
    TRANSMITTERS = None
    EMITTERS = None
    TRANSMITTERS_AND_EMITTERS = None


class UnitType(MutableEnum):
    POWER = None
    FREQUENCY = None
    LENGTH = None
    TIME = None
    VOLTAGE = None
    DATA_RATE = None
    RESISTANCE = None


EMIT_VALID_UNITS = {
    "Power": ["mW", "W", "kW", "dBm", "dBW"],
    "Frequency": ["Hz", "kHz", "MHz", "GHz", "THz"],
    "Length": ["pm", "nm", "um", "mm", "cm", "dm", "meter", "km", "mil", "in", "ft", "yd", "mile"],
    "Time": ["ps", "ns", "us", "ms", "s"],
    "Voltage": ["mV", "V"],
    "Data Rate": ["bps", "kbps", "Mbps", "Gbps"],
    "Resistance": ["uOhm", "mOhm", "Ohm", "kOhm", "megOhm", "GOhm"],
}
"""Valid units for each unit type."""


def emit_unit_type_string_to_enum(unit_string):
    EMIT_UNIT_TYPE_STRING_TO_ENUM = {
        "Power": UnitType.POWER,
        "Frequency": UnitType.FREQUENCY,
        "Length": UnitType.LENGTH,
        "Time": UnitType.TIME,
        "Voltage": UnitType.VOLTAGE,
        "Data Rate": UnitType.DATA_RATE,
        "Resistance": UnitType.RESISTANCE,
    }
    return EMIT_UNIT_TYPE_STRING_TO_ENUM[unit_string]


class EmiCategoryFilter(MutableEnum):
    IN_CHANNEL_TX_FUNDAMENTAL = None
    IN_CHANNEL_TX_HARMONIC_SPURIOUS = None
    IN_CHANNEL_TX_INTERMOD = None
    IN_CHANNEL_TX_BROADBAND = None
    OUT_OF_CHANNEL_TX_FUNDAMENTAL = None
    OUT_OF_CHANNEL_TX_HARMONIC_SPURIOUS = None
    OUT_OF_CHANNEL_TX_INTERMOD = None


if __name__ == "__main__":
    print("Members of EmiCategoryFilter:")
    for m in EmiCategoryFilter.members():
        print("    {}".format(m))



import argparse
import os
import shutil
import sys
import warnings
from xml.dom.minidom import parseString
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ParseError

current_dir = os.path.dirname(os.path.realpath(__file__))
pyaedt_path = os.path.normpath(
    os.path.join(
        current_dir,
        "..",
    )
)
sys.path.append(os.path.normpath(os.path.join(pyaedt_path, "..")))

is_linux = os.name == "posix"
is_windows = not is_linux
pid = 0


def main():
    args = parse_arguments()
    add_pyaedt_to_aedt(
        args.version, is_student_version=args.student, use_sys_lib=args.sys_lib, new_desktop_session=args.new_session
    )


def parse_arguments():
    parser = argparse.ArgumentParser(description="Install PyAEDT and setup PyAEDT toolkits in AEDT.")
    parser.add_argument(
        "--version", "-v", default="231", metavar="XY.Z", help="AEDT three-digit version (e.g. 231). Default=231"
    )
    parser.add_argument(
        "--student", "--student_version", action="store_true", help="Install toolkits for AEDT Student Version."
    )
    parser.add_argument("--sys_lib", "--syslib", action="store_true", help="Install toolkits in SysLib.")
    parser.add_argument(
        "--new_session", action="store_true", help="Start a new session of AEDT after installing PyAEDT."
    )

    args = parser.parse_args()
    args = process_arguments(args, parser)
    return args


def process_arguments(args, parser):
    if len(args.version) != 3:
        parser.print_help()
        parser.error("Version should be a three digit number (e.g. 231)")

    args.version = "20" + args.version[-3:-1] + "." + args.version[-1:]
    return args


def add_pyaedt_to_aedt(
    aedt_version, is_student_version=False, use_sys_lib=False, new_desktop_session=False, sys_dir="", pers_dir=""
):
    if not (sys_dir or pers_dir):
        from pyaedt import Desktop
        from pyaedt.generic.general_methods import grpc_active_sessions
        from pyaedt.generic.settings import settings

        sessions = grpc_active_sessions(aedt_version, is_student_version)
        close_on_exit = True
        if not sessions:
            if not new_desktop_session:
                print("Launching a new AEDT desktop session.")
            new_desktop_session = True
        else:
            close_on_exit = False
        settings.use_grpc_api = True
        with Desktop(
            specified_version=aedt_version,
            non_graphical=new_desktop_session,
            new_desktop_session=new_desktop_session,
            student_version=is_student_version,
            close_on_exit=close_on_exit,
        ) as d:
            desktop = sys.modules["__main__"].oDesktop
            pers1 = os.path.join(desktop.GetPersonalLibDirectory(), "pyaedt")
            pid = desktop.GetProcessID()
            # Linking pyaedt in PersonalLib for IronPython compatibility.
            if os.path.exists(pers1):
                d.logger.info("PersonalLib already mapped.")
            else:
                if is_windows:
                    os.system('mklink /D "{}" "{}"'.format(pers1, pyaedt_path))
                else:
                    os.system('ln -s "{}" "{}"'.format(pyaedt_path, pers1))
            sys_dir = d.syslib
            pers_dir = d.personallib
        if pid and new_desktop_session:
            try:
                os.kill(pid, 9)
            except:
                pass

    toolkits = ["Project"]
    # Bug on Linux 23.1 and before where Project level toolkits don't show up. Thus copying to individual design
    # toolkits.
    if is_linux and aedt_version <= "2023.1":
        toolkits = [
            "2DExtractor",
            "CircuitDesign",
            "HFSS",
            "HFSS-IE",
            "HFSS3DLayoutDesign",
            "Icepak",
            "Maxwell2D",
            "Maxwell3D",
            "Q3DExtractor",
            "Mechanical",
        ]

    for product in toolkits:
        if use_sys_lib:
            try:
                sys_dir = os.path.join(sys_dir, "Toolkits")
                install_toolkit(sys_dir, product, aedt_version)
                print("Installed toolkit for {} in sys lib.".format(product))
                # d.logger.info("Installed toolkit for {} in sys lib.".format(product))

            except IOError:
                pers_dir = os.path.join(pers_dir, "Toolkits")
                install_toolkit(pers_dir, product, aedt_version)
                print("Installed toolkit for {} in sys lib.".format(product))
                # d.logger.info("Installed toolkit for {} in personal lib.".format(product))
        else:
            pers_dir = os.path.join(pers_dir, "Toolkits")
            install_toolkit(pers_dir, product, aedt_version)
            print("Installed toolkit for {} in sys lib.".format(product))
            # d.logger.info("Installed toolkit for {} in personal lib.".format(product))


def install_toolkit(toolkit_dir, product, aedt_version):
    tool_dir = os.path.join(toolkit_dir, product, "PyAEDT")
    lib_dir = os.path.join(tool_dir, "Lib")
    toolkit_rel_lib_dir = os.path.relpath(lib_dir, tool_dir)
    # Bug on Linux 23.1 and before where Project level toolkits don't show up. Thus copying to individual design
    # toolkits.
    if is_linux and aedt_version <= "2023.1":
        toolkit_rel_lib_dir = os.path.join("Lib", "PyAEDT")
        lib_dir = os.path.join(toolkit_dir, toolkit_rel_lib_dir)
        toolkit_rel_lib_dir = "../../" + toolkit_rel_lib_dir
        tool_dir = os.path.join(toolkit_dir, product, "PyAEDT")
    os.makedirs(lib_dir, exist_ok=True)
    os.makedirs(tool_dir, exist_ok=True)
    files_to_copy = ["Console", "Run_PyAEDT_Script", "Jupyter"]
    # Remove hard-coded version number from Python virtual environment path, and replace it with the corresponding AEDT
    # version's Python virtual environment.
    version_agnostic = False
    if aedt_version[2:6].replace(".", "") in sys.executable:
        executable_version_agnostic = sys.executable.replace(aedt_version[2:6].replace(".", ""), "%s")
        version_agnostic = True
    else:
        executable_version_agnostic = sys.executable
    jupyter_executable = executable_version_agnostic.replace("python" + exe(), "jupyter" + exe())
    ipython_executable = executable_version_agnostic.replace("python" + exe(), "ipython" + exe())
    for file_name in files_to_copy:
        with open(os.path.join(current_dir, file_name + ".py_build"), "r") as build_file:
            file_name_dest = file_name.replace("_", " ") + ".py"
            with open(os.path.join(tool_dir, file_name_dest), "w") as out_file:
                print("Building to " + os.path.join(tool_dir, file_name_dest))
                build_file_data = build_file.read()
                build_file_data = (
                    build_file_data.replace("##TOOLKIT_REL_LIB_DIR##", toolkit_rel_lib_dir)
                    .replace("##PYTHON_EXE##", executable_version_agnostic)
                    .replace("##IPYTHON_EXE##", ipython_executable)
                    .replace("##JUPYTER_EXE##", jupyter_executable)
                )
                if not version_agnostic:
                    build_file_data = build_file_data.replace(" % version", "")
                out_file.write(build_file_data)
    shutil.copyfile(os.path.join(current_dir, "console_setup.py"), os.path.join(lib_dir, "console_setup.py"))
    shutil.copyfile(
        os.path.join(current_dir, "jupyter_template.ipynb"),
        os.path.join(lib_dir, "jupyter_template.ipynb"),
    )
    if aedt_version >= "2023.2":
        write_tab_config(os.path.join(toolkit_dir, product), lib_dir)


def write_tab_config(product_toolkit_dir, pyaedt_lib_dir, force_write=False):
    tab_config_file_path = os.path.join(product_toolkit_dir, "TabConfig.xml")
    if not os.path.isfile(tab_config_file_path) or force_write:
        root = ET.Element("TabConfig")
    else:
        try:
            tree = ET.parse(tab_config_file_path)
        except ParseError as e:
            warnings.warn("Unable to parse %s\nError received = %s" % (tab_config_file_path, str(e)))
            return
        root = tree.getroot()

    panels = root.findall("./panel")
    if panels:
        panel_names = [panel.attrib["label"] for panel in panels]
        if "Panel_PyAEDT" in panel_names:
            # Remove previously existing PyAEDT panel and update with newer one.
            panel = [panel for panel in panels if panel.attrib["label"] == "Panel_PyAEDT"][0]
            root.remove(panel)

    # Write a new "Panel_PyAEDT" sub-element.
    panel = ET.SubElement(root, "panel", label="Panel_PyAEDT")
    gallery = ET.SubElement(panel, "gallery", imagewidth="120", imageheight="72")
    image_rel_path = os.path.relpath(pyaedt_lib_dir, product_toolkit_dir).replace("\\", "/") + "/"
    if image_rel_path == "./":
        image_rel_path = ""
    ET.SubElement(gallery, "button", label="PyAEDT", isLarge="1", image=image_rel_path + "images/large/pyansys.png")
    group = ET.SubElement(gallery, "group", label="PyAEDT Menu", image=image_rel_path + "images/gallery/PyAEDT.png")
    ET.SubElement(group, "button", label="Console", script="PyAEDT/Console")
    ET.SubElement(group, "button", label="Jupyter Notebook", script="PyAEDT/Jupyter")
    ET.SubElement(group, "button", label="Run PyAEDT Script", script="PyAEDT/Run PyAEDT Script")

    # Backup any existing file if present
    if os.path.isfile(tab_config_file_path):
        shutil.copy(tab_config_file_path, tab_config_file_path + ".orig")

    write_pretty_xml(root, tab_config_file_path)

    files_to_copy = ["images/large/pyansys.png", "images/gallery/PyAEDT.png"]
    for file_name in files_to_copy:
        dest_file = os.path.normpath(os.path.join(pyaedt_lib_dir, file_name))
        os.makedirs(os.path.dirname(dest_file), exist_ok=True)
        shutil.copy(os.path.normpath(os.path.join(current_dir, file_name)), dest_file)


def write_pretty_xml(root, file_path):
    """Write the XML in a pretty format."""
    # If we use the commented code below, then the previously existing lines will have double lines added. We need to
    # split and ignore the double lines.
    # xml_str = parseString(ET.tostring(root)).toprettyxml(indent=" " * 4)
    lines = [line for line in parseString(ET.tostring(root)).toprettyxml(indent=" " * 4).split("\n") if line.strip()]
    xml_str = "\n".join(lines)

    with open(file_path, "w") as f:
        f.write(xml_str)


def exe():
    if is_windows:
        return ".exe"
    return ""


if __name__ == "__main__":
    main()



import re
import types

from pyaedt.generic.general_methods import GrpcApiError
from pyaedt.generic.general_methods import _retry_ntimes
from pyaedt.generic.general_methods import settings
import pyaedt.generic.grpc_plugin_dll as AedtAPI

logger = settings.logger
__all__ = ["CreateAedtApplication", "Release"]


def CreateAedtObj(objectID, bIsPropSvr, listFuncs):
    # print("Create " + str(objectID))
    if bIsPropSvr:
        return AedtPropServer(objectID, listFuncs)
    return AedtObjWrapper(objectID, listFuncs)


def CreateAedtBlockObj(list):
    count = len(list)
    if count > 1:
        if isinstance(list[0], str):
            toks = list[0].split(":")
            if len(toks) == 2:
                start = -1
                if count % 2 == 0:
                    if toks[1] == "=":
                        start = 2
                elif count > 2:
                    if toks[0] == "NAME":
                        start = 1
                if start > 0:
                    isBlock = True
                    for i in range(start, count - 1, 2):
                        if isinstance(list[i], str):
                            toks = list[i].split(":")
                            if len(toks) != 2 or toks[1] != "=":
                                isBlock = False
                                break
                        else:
                            isBlock = False
                            break
                    if isBlock:
                        return AedtBlockObj(list)
    return list


def GetAedtObjId(obj):
    if isinstance(obj, AedtObjWrapper):
        return obj.objectID
    return None


class AedtBlockObj(list):
    def GetName(self):
        if len(self) > 0:
            f = self[0]
            if isinstance(f, str):
                d = f.split(":")
                if len(d) == 2:
                    if d[0] == "NAME":
                        return d[1]

    def __GetValueIdxByKey__(self, keyName):
        for i in range(0, len(self) - 1):
            if isinstance(self[i], str):
                toks = self[i].split(":")
                if len(toks) == 2:
                    if toks[1] == "=" and toks[0] == keyName:
                        return i + 1
        raise GrpcApiError(keyName + " is not a key!")

    def __getitem__(self, idxOrKey):
        if isinstance(idxOrKey, str):
            idx = self.__GetValueIdxByKey__(idxOrKey)
            if idx != None:
                return super().__getitem__(idx)
        return super().__getitem__(idxOrKey)

    def __setitem__(self, idxOrKey, newVal):
        if isinstance(idxOrKey, int):
            if idxOrKey >= 0 or idxOrKey < len(self):
                oldItem = self.__getitem__(idxOrKey)
                if isinstance(oldItem, str):
                    toks = oldItem.split(":")
                    if len(toks) == 2 and (toks[1] == "=" or toks[0] == "NAME"):
                        raise GrpcApiError("The element is a key should not be overwritten!")
            return super().__setitem__(idxOrKey, newVal)
        if isinstance(idxOrKey, str):
            idx = self.__GetValueIdxByKey__(idxOrKey)
            if idx != None:
                return super().__setitem__(idx, newVal)
            raise GrpcApiError("Key not found")
        raise GrpcApiError("Must be key name or index")

    def keys(self):
        arr = []
        for i in range(0, len(self) - 1):
            if isinstance(self[i], str):
                toks = self[i].split(":")
                if len(toks) == 2 and toks[1] == "=":
                    arr.append(toks[0])
        return arr


class AedtObjWrapper:
    def __init__(self, objID, listFuncs):
        self.__dict__["objectID"] = objID  # avoid derive class overwrite __setattr__
        self.__dict__["__methodNames__"] = listFuncs
        # print(self.objectID)

    def __str__(self):
        return "Instance of an Aedt object:" + str(self.objectID)

    def __Invoke__(self, funcName, argv):
        if settings.enable_debug_grpc_api_logger:
            settings.logger.debug("{} {}".format(funcName, argv))
        try:
            return _retry_ntimes(
                settings.number_of_grpc_api_retries, AedtAPI.InvokeAedtObjMethod, self.objectID, funcName, argv
            )  # Call C function
        except:  # pragma: no cover
            raise GrpcApiError("Failed to execute grpc AEDT command: {}".format(funcName))

    def __dir__(self):
        return self.__methodNames__

    def __GetObjMethod__(self, funcName):
        try:

            def DynamicFunc(self, *args):
                return self.__Invoke__(funcName, args)

            return types.MethodType(DynamicFunc, self)
        except (AttributeError, GrpcApiError):
            raise GrpcApiError("This Aedt object has no attribute '" + funcName + "'")

    def __getattr__(self, funcName):
        try:
            if funcName == "ScopeID":  # backward compatible for IronPython wrapper.
                return self.objectID
            return self.__GetObjMethod__(funcName)
        except:
            raise GrpcApiError("Failed to get grpc API AEDT attribute {}".format(funcName))

    def __setattr__(self, attrName, val):
        if attrName == "objectID" or attrName == "__methodNames__":
            raise GrpcApiError("Modify this attribute is not allowed")
        elif attrName in self.__methodNames__:
            raise GrpcApiError(attrName + " is a function name")
        else:
            super().__setattr__(attrName, val)

    def __del__(self):
        AedtAPI.ReleaseAedtObject(self.objectID)

    def match(self, patternStr):  # IronPython wrapper implemented this function return IEnumerable<string>.
        class IEnumerable(list):
            def __getattr__(self, key):
                if key == "Count":
                    return len(self)

        pattern = re.compile(patternStr)
        found = IEnumerable()
        allMethods = self.__methodNames__
        for method in allMethods:
            if pattern.match(method):
                found.append(method)
        return found

    def GetHashCode(self):  # IronPython build in function
        return self.__hash__()


class AedtPropServer(AedtObjWrapper):
    def __init__(self, objID, listFuncs):
        AedtObjWrapper.__init__(self, objID, listFuncs)
        self.__dict__["__propMap__"] = None
        self.__dict__["__propNames__"] = None

    def __GetPropAttributes(self):
        if self.__propMap__ == None:
            propMap = {}
            propNames = self.GetPropNames()
            for prop in propNames:
                attrName = ""
                if prop[0].isdigit():
                    attrName += "_"
                for c in prop:
                    if c.isalnum() == True:
                        attrName += c
                    else:
                        attrName += "_"
                propMap[attrName] = prop
            self.__propMap__ = propMap
        return self.__propMap__

    def __dir__(self):
        ret = super().__dir__().copy()
        for attrName in self.__GetPropAttributes().keys():
            ret.append(attrName)
        return ret

    def __getattr__(self, attrName):
        try:
            return super().__getattr__(attrName)
        except AttributeError:
            # if AedtAPI.IsAedtObjPropName(self.objectID, attrName, False):
            #    return self.GetPropValue(attrName)
            propMap = self.__GetPropAttributes()
            if attrName in propMap:
                return self.GetPropValue(propMap[attrName])
            raise GrpcApiError("Failed to retrieve attribute {} from GRPC API".format(attrName))

    def __setattr__(self, attrName, val):
        if attrName in self.__dict__:
            self.__dict__[attrName] = val
            return

        # if AedtAPI.IsAedtObjPropName(self.objectID, attrName, True):
        #    self.SetPropValue(attrName, val)
        #    return
        propMap = self.__GetPropAttributes()
        if attrName in propMap:
            self.SetPropValue(propMap[attrName], val)
            return
        super().__setattr__(attrName, val)

    def GetName(self):
        return self.__Invoke__("GetName", ())

    def GetObjPath(self):
        return self.__Invoke__("GetObjPath", ())

    def GetChildNames(self, childType=""):
        return self.__Invoke__("GetChildNames", (childType))

    def GetPropNames(self, includeReadOnly=True):
        if includeReadOnly:
            if self.__propNames__ == None:
                self.__propNames__ = self.__Invoke__("GetPropNames", (includeReadOnly,))
            return self.__propNames__
        return self.__Invoke__("GetPropNames", (includeReadOnly,))

    def GetPropValue(self, propName=""):
        return self.__Invoke__("GetPropValue", (propName,))

    def SetPropValue(self, propName, val):
        return self.__Invoke__("SetPropValue", (propName, val))


def CreateAedtApplication(machine="", port=0, NGmode=False, alwaysNew=True):
    return AedtAPI.CreateAedtApplication(machine, port, NGmode, alwaysNew)


def Release():
    AedtAPI.ReleaseAll()


AedtAPI.SetPyObjCalbacks(CreateAedtObj, CreateAedtBlockObj, GetAedtObjId)



import warnings

from pyaedt import emit_core
from pyaedt.emit_core.results.revision import Revision
from pyaedt.generic.general_methods import pyaedt_function_handler


class Results:
    """
    Provides the ``Results`` object.

    Parameters
    ----------
    emit_obj : emit_obj object
        EMIT object used to create the result.

    Examples
    --------
    Create an instance of the ``Result`` object.

    >>> aedtapp.results = Results()
    >>> revision = aedtapp.results.analyze()
    >>> receivers = revision.get_receiver_names()
    """

    def __init__(self, emit_obj):
        self.emit_project = emit_obj
        """EMIT project."""

        self.current_revision = None
        """Current active Revision."""

        self.revisions = []
        """List of all result revisions. Only one loaded at a time"""

        self.design = emit_obj.odesktop.GetActiveProject().GetActiveDesign()
        """Active design for the EMIT project."""

    @pyaedt_function_handler()
    def _add_revision(self, name=None):
        """Add a new revision or get the current revision if it already exists.

        Parameters
        ----------
        name : str, optional
            Name for the new revision, if created. The default is ``None``, in which
            case the name of the current design revision is used.

        Raises
        ------
        RuntimeError if the name given is not the name of an existing result set and a current result set already
        exists.

        Returns
        -------
        ``Revision`` object that was created.
        """
        revision = Revision(self, self.emit_project, name)
        self.revisions.append(revision)
        return revision

    @pyaedt_function_handler()
    def delete_revision(self, revision_name):
        """Delete the specified revision from the results.

        Parameters
        ----------
        revision_name : str
            Name of the revision.

        Returns
        -------
        None

        Examples
        --------
        >>> aedtapp.results.delete_revision("Revision 10")
        """
        if revision_name in self.design.GetResultList():
            self.design.DeleteResult(revision_name)
            if self.current_revision.name == revision_name and self.current_revision.revision_loaded:
                self.emit_project._emit_api.close()
                self.current_revision = None
            for rev in self.revisions:
                if revision_name in rev.name:
                    self.revisions.remove(rev)
                    break
            else:
                warnings.warn("{} does not exist".format(revision_name))

    @staticmethod
    def interaction_domain():
        """
        Get an ``InteractionDomain`` object.

        Returns
        -------
        :class:`Emit.InteractionDomain`
            Defines a set of interacting interferers and receivers.

        Examples
        --------
        >>> domain = Emit.results.InteractionDomain()

        """
        try:
            domain = emit_core.emit_api_python().InteractionDomain()
        except NameError:
            raise ValueError("An EMIT object must be initialized before any static member of the Results.")
        return domain

    @pyaedt_function_handler
    def _unload_revisions(self):
        """Convenience function to set all revisions
        as ``unloaded``

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        for rev in self.revisions:
            rev.revision_loaded = False

    @pyaedt_function_handler()
    def revision_names(self):
        """
        Return a list of all the revision names.

        Parameters
        ----------
        None

        Returns
        -------
        revision_names : list str
            List of all revision names.
        """
        return [rev.name for rev in self.revisions]

    @pyaedt_function_handler
    def get_revision(self, revision_name=None):
        """
        Load the specified revision.

        Parameters
        ----------
        revision_name : str, optional
            Revision to load. The default is  ``None`` in which
            case the latest revision will be returned.

        Returns
        -------
        rev:class:`pyaedt.modules.Revision`
            Specified ``Revision`` object that was loaded.

        Examples
        --------
        >>> rev = aedtapp.results.get_revision("Revision 15")
        >>> interferers = rev.get_interferer_names()
        >>> receivers = rev.get_receiver_names()
        """
        # no revisions to load, create a new one
        if len(self.revisions) == 0:
            return self.analyze()
        # retrieve the latest revision if nothing specified
        if revision_name is None:
            # unload the current revision and load the latest
            self.current_revision.revision_loaded = False
            self.current_revision = self.revisions[-1]
            self.current_revision._load_revision()
        else:
            rev = [x for x in self.revisions if revision_name == x.name]
            if len(rev) > 0:
                # unload the current revision and load the specified revision
                self.current_revision.revision_loaded = False
                self.current_revision = rev[0]
                self.current_revision._load_revision()
            else:
                # might be an old revision that was never loaded by pyaedt
                aedt_result_list = self.design.GetResultList()
                rev = [x for x in aedt_result_list if revision_name == x]
                if len(rev) > 0:
                    # unload the current revision and load the specified revision
                    self.current_revision.revision_loaded = False
                    self.current_revision = self._add_revision(rev[0])
                else:
                    warnings.warn("{} not found.".format(revision_name))
        return self.current_revision

    @pyaedt_function_handler()
    def analyze(self):
        """
        Analyze the current revision or create a new revision if
        the design has changed.

        Returns
        -------
        rev:class:`pyaedt.modules.Revision`
            Specified ``Revision`` object that was generated.

        Examples
        --------
        >>> rev = aedtapp.results.analyze()
        >>> interferers = rev.get_interferer_names()
        >>> receivers = rev.get_receiver_names()
        """
        # No revisions exist, add one
        if self.current_revision is None:
            self.current_revision = self._add_revision()
        # no changes since last created revision, load it
        elif (
            self.revisions[-1].revision_number
            == self.emit_project.odesktop.GetActiveProject().GetActiveDesign().GetRevision()
        ):
            self.get_revision(self.revisions[-1].name)
        else:
            # there are changes since the current revision was analyzed, create
            # a new revision
            self.current_revision.revision_loaded = False
            self.current_revision = self._add_revision()

        return self.current_revision
